@@ -14,8 +14,8 @@
 :Copyright: GNU Public License v2
 
 This module provides methods for finding neighbouring lipids in a bilayer,
-calculating local lipid compositions and finding the largest cluster of
-lipids over time.
+calculating local lipid compositions and lipid enrichment, and finding the
+largest cluster of specific species of lipids over time.
 
 Two lipids are considered neighbours if they have any atoms within a given
 cutoff distance of one another.
@@ -37,20 +37,14 @@ Output
   - *neighbours* : a sparse matrix of binary variables, equal to 1 if two lipids are in contact, and 0 otherwise
 
 For efficient use of memory, an adjacency matrix of neighbouring lipids is stored
-in a :class:`scipy.sparse.csc_matrix` sparse matrix for each frame of the analysis. The data
-are stored in the :attr:`neighbours.neighbours` attribute. The matrix has shape
-(n_residues, n_residues * n_frames), and the data for frame *n* is accessed via::
-
-  neighbours.neighbours[:, (n * n_residues):((n + 1) * n_residues)]
-
-A matrix element *[i, j]* is equal to 1 if lipid *i* neighbours lipid *j* and equal to 0 otherwise.
-The matrix is symmetric: if lipid *i* neighbours lipid *j* then *j* must neighbour *i*.
-
+in a :class:`scipy.sparse.csr_matrix` sparse matrix for each frame of the analysis. The data
+are stored in the :attr:`neighbours.neighbours` attribute as a NumPy array of sparse
+matrices. Each matrix has shape (n_residues, n_residues)
 
 Tip
 ---
 
-The resultant sparse matrix can be used to calculate the loca lipid composition of each individual lipid
+The resultant sparse matrix can be used to calculate the local lipid composition of each individual lipid
 at each frame using :func:`lipyphilic.lib.neighbours.count_neighbours`, or to find the largest cluster of
 lipids at each frame using :func:`lipyphilic.lib.neighbours.largest_cluster`.
 
@@ -88,7 +82,7 @@ frame) and select to display a progress bar (`verbose=True`)::
   )
   
 The results are then available in the :attr:`neighbours.Neighbours` attribute as a
-:class:`scipy.sparse.csc_matrix`.
+:class:`numpy.ndarray` of Compressed Sparse Row matrices.
 
 Counting the number of neighbours: by lipid species
 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
@@ -142,6 +136,27 @@ in each row::
         <total num neighbours>
     ]
 
+Calculate the enrichment index of lipid species
+^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+
+The :func:`count_neighbours` method will, by default, return the number of neighbouring lipids
+around each individual lipid.
+
+However, a clearer picture of aggregation of certain lipid species can
+be gained by instead considering the enrichment/depletion index of each lipid species, defined in
+`Ingólfsson et al. (2014) <https://pubs.acs.org/doi/10.1021/ja507832e>`__. In this
+instance, the number of each neighbour species B around a given reference species A is normalized
+by the average number of species B around any lipid.
+
+To calculate the enrichment/depletion index of each species at each frame, as well as the raw
+neighbour counts, we can set the :attr:`return_enrichment` keyword to true::
+
+  counts, enrichment = neighbours.count_neighbours(return_enrichment=True)
+
+This will return two :mod:`pandas` :class:`DataFrames`, one containing the neighbour counts
+and the other the enrichment/depletion index of each species at each frame. The benefit of having
+the enrichment index at each frame is that you can plot its time-evolution to see whether
+particular species form aggregates over time.
 
 Find the largest cluster
 ^^^^^^^^^^^^^^^^^^^^^^^^
@@ -189,7 +204,7 @@ where `True` indicates that we should include this lipid in the neighbour calcul
 
   largest_cluster_upper_leaflet = neighbours.largest_cluster(
     cluster_sel="resname CHOL DPPC",
-    leaflets=upper_leaflet_mask
+    filter_by=upper_leaflet_mask
   )
 
 Now, lipids either in the lower leaflet (-1) or the midplane (0) will not be included when determining
@@ -269,11 +284,7 @@ class Neighbours(base.AnalysisBase):
         
     def _prepare(self):
         
-        # Output array
-        self.neighbours = scipy.sparse.lil_matrix(
-            (self.membrane.n_residues, self.membrane.n_residues * self.n_frames),
-            dtype=np.int8
-        )
+        self.neighbours = np.zeros(self.n_frames, dtype=object)
         
     def _single_frame(self):
         
@@ -284,7 +295,6 @@ class Neighbours(base.AnalysisBase):
             box=self._ts.dimensions,
             return_distances=False
         )
-        self.pairs = pairs
         
         # Find unique pairs of residues interacting
         # Currently we have pairs of atoms
@@ -296,17 +306,14 @@ class Neighbours(base.AnalysisBase):
         neigh = neigh[different]
         
         # store neighbours for this frame
-        frame_start = self._frame_index * self.membrane.n_residues
-        self.neighbours[ref, neigh + frame_start] = 1
-        self.neighbours[neigh, ref + frame_start] = 1  # the neighbour matrix must be symmetric
-
-    def _conclude(self):
-        
-        # csc sparse matrices are faster for slicing columns
-        # we'll want this for selecting neighbours at given frames
-        self.neighbours = self.neighbours.tocsc()
-
-    def count_neighbours(self, count_by=None, count_by_labels=None):
+        data = np.ones_like(ref)
+        self.neighbours[self._frame_index] = scipy.sparse.csr_matrix(
+            (data, (ref, neigh)),
+            dtype=np.int8,
+            shape=(self.membrane.n_residues, self.membrane.n_residues)
+        )
+    
+    def count_neighbours(self, count_by=None, count_by_labels=None, return_enrichment=False):
         """Count the number of each neighbour type at each frame.
 
         Parameters
@@ -322,6 +329,10 @@ class Neighbours(base.AnalysisBase):
             `count_by_labels = {'Ld': 0, 'Lo': 1}`. There **must** be precisely one label for each unique
             value in 'count_by'. If `count_by` is given but `count_by_labels` is left as `None`, the values
             in `count_by` will be used as the labels.
+        return_enrichment : bool, optional
+            If `True`, a second DataFrame containing the fractional enrichment of each lipid species at each
+            frame is also returned. The default is `False`, in which case the fractional enrichment
+            if not returned.
         
         Returns
         -------
@@ -331,6 +342,10 @@ class Neighbours(base.AnalysisBase):
             (default is resname), lipid residue index, frame number, number of neighbours of each species
             (or of each type in 'count_by' if this is provided), as well as the total number of neighbours.
         
+        enrichment : pandas.DataFrame
+            A DataFrame containing the following data enrichment/depletion data for each lipid species at
+            each frame.
+        
         """
         
         if self.neighbours is None:
@@ -373,9 +388,9 @@ class Neighbours(base.AnalysisBase):
         
         # Get counts at each frame
         n_residues = self.membrane.n_residues
-        for frame_index in tqdm(np.arange(self.n_frames)):
+        for frame_index, neighbours in tqdm(enumerate(self.neighbours), total=self.n_frames):
         
-            ref, neigh = self.neighbours[:, frame_index * n_residues:(frame_index + 1) * n_residues].nonzero()
+            ref, neigh = neighbours.nonzero()
             unique, counts = np.unique([ref, [type_index[t] for t in count_by[neigh, frame_index]]], axis=1, return_counts=True)
             
             r, t = unique  # reference index (r) and type index (t)
@@ -393,23 +408,61 @@ class Neighbours(base.AnalysisBase):
         all_counts = all_counts.reshape(n_residues * self.n_frames, len(count_by_labels))
         total_counts = np.sum(all_counts, axis=1)
         
-        data = np.concatenate(
-            (labels[:, np.newaxis], resindices[:, np.newaxis], frames[:, np.newaxis], all_counts, total_counts[:, np.newaxis]),
-            axis=1
-            )
-        
-        # Create DataFrame
-        columns = ["Label", "Resindex", "Frame"] + [f"n{label}" for label in count_by_labels] + ["Total"]
-        df = pd.DataFrame(
-            data=data,
-            columns=columns
+        # Create the dataframe
+        counts = pd.DataFrame(
+            data=labels,
+            columns=["Label"]
         )
+
+        counts["Resindex"] = resindices
+        counts["Frame"] = frames
+
+        for count_by_label in count_by_labels:
+            counts[f"n{count_by_label}"] = all_counts.T[type_index[count_by_label]]
+
+        counts["Total"] = total_counts
         
         # make every column except the label take on integer values
-        for column in df.columns[1:]:
-            df[column] = pd.to_numeric(df[column])
+        for column in counts.columns[1:]:
+            counts[column] = pd.to_numeric(counts[column])
         
-        return df
+        if return_enrichment is False:
+            return counts
+        
+        # Otherwise create a second DataFrame containing the fractional enrichment
+        unique_labels = [label for label in type_index]
+
+        # We need to normalize the count by the mean number of neighbours of each species
+        mean_neighbours_counts = np.asarray(
+            [counts.groupby("Frame")[neigh].mean().values for neigh in [f"n{label}" for label in unique_labels]]
+        )
+        n_unique_labels, n_frames = mean_neighbours_counts.shape
+        
+        # create new output arrays
+        labels = np.full((n_frames, n_unique_labels), fill_value=unique_labels).T.flatten()
+        neighbour_enrichment = np.full((n_frames * n_unique_labels, n_unique_labels), fill_value=np.NaN)
+        
+        # and the new DataFrame
+        enrichment = pd.DataFrame(
+            data=labels,
+            columns=["Label"]
+        )
+        enrichment["Frame"] = np.full((n_unique_labels, n_frames), fill_value=counts["Frame"].unique()).flatten()
+        
+        # Calculate the enrichment of each species at each frame
+        for species_index, ref in enumerate(unique_labels):
+        
+            ref_mask = (counts.Label == ref).values
+            
+            species_neighbour_counts = counts.loc[ref_mask]
+            species_neighbour_enrichment = species_neighbour_counts.groupby("Frame")[[f"n{label}" for label in unique_labels]].mean() / mean_neighbours_counts.T
+            neighbour_enrichment[n_frames * species_index:n_frames * (species_index + 1)] = species_neighbour_enrichment
+            
+        # Finally add the enrichment values to the DataFrame
+        for species_index, ref in enumerate([f"fe{label}" for label in unique_labels]):
+            enrichment[ref] = neighbour_enrichment[:, species_index]
+        
+        return counts, enrichment
     
     def largest_cluster(self, cluster_sel=None, filter_by=None, return_indices=False):
         """Find the largest cluster of lipids at each frame.
@@ -433,13 +486,15 @@ class Neighbours(base.AnalysisBase):
             there are two largest clusters of equal size, only the residue indices of lipids in one
             cluster will be returned (the cluster that has the lipid with the smallest residue index). The
             default is `False`, in which case no reidue indices are returned.
-            
         
         Returns
         -------
         
         largest_cluster : numpy.ndarray
             An array containing the number of lipids in the largest cluster at each frame.
+        indices : list
+            A list of 1D NumPy arrays, where each array corresponds to a single frame and contains the
+            residue indices of lipids in the largest cluster at that frame.
             
         Note
         ----
@@ -505,14 +560,11 @@ class Neighbours(base.AnalysisBase):
         # output arrays
         largest_cluster = np.zeros(self.n_frames, dtype=int)
         largest_cluster_resindices = np.full(self.n_frames, fill_value=0, dtype=object)
-
-        n_residues = self.membrane.n_residues
-        for frame_index, frame in enumerate(self.frames):
+        
+        for frame_index, neighbours in tqdm(enumerate(self.neighbours), total=self.n_frames):
             
             frame_filter = filter_by[:, frame_index]
-            
-            frame_neighbours = self.neighbours[:, frame_index * n_residues:(frame_index + 1) * n_residues]
-            frame_neighbours = frame_neighbours[frame_filter][:, frame_filter].toarray()
+            frame_neighbours = neighbours[frame_filter][:, frame_filter]
             
             # find all connected components
             _, com_labels = scipy.sparse.csgraph.connected_components(frame_neighbours)
